import difflib
import os.path

__all__ = ["unified_diff"]


def unified_diff(
    astr: str,
    bstr: str,
    filename: str,
    n: int = 3,
) -> str:
    """
    Returns a unified diff string for the two inputs.

    Does not currently support creation or deletion where one of the filenames
    is `/dev/null` or patchlevels other than `-p1`.

    Does handle the "no newline at end of file" properly UNLIKE DIFFLIB.
    """
    if os.path.isabs(filename):
        a_filename = filename
        b_filename = filename
    else:
        a_filename = f"a/{filename}"
        b_filename = f"b/{filename}"

    buf = []
    gen = difflib.unified_diff(
        astr.splitlines(True),
        bstr.splitlines(True),
        a_filename,
        b_filename,
        n=n,
    )
    for line in gen:
        buf.append(line)
        if not line.endswith("\n"):
            # Assume this is the only case where it can happen
            buf.append("\n\\ No newline at end of file\n")

    return "".join(buf)



from unittest import TestCase

from moreorless.combined import _contributions, combined_diff
from moreorless.combined_types import Hunk, Line


class LineTest(TestCase):
    def test_common_only(self) -> None:
        li = Line((0, 0, 0, 1), "foo\n")
        self.assertFalse(li.is_context)
        self.assertEqual("+++foo\n", li.to_str(is_merge=True))
        self.assertEqual("---foo\n", li.to_str(is_merge=False))

    def test_context_only(self) -> None:
        li = Line((1, 1, 1, 1), "foo\n")
        self.assertTrue(li.is_context)
        self.assertEqual("   foo\n", li.to_str(is_merge=True))
        self.assertEqual("   foo\n", li.to_str(is_merge=False))

    def test_mixed(self) -> None:
        li = Line((1, 1, 0, 0), "foo\n")
        self.assertFalse(li.is_context)
        self.assertEqual("-- foo\n", li.to_str(is_merge=True))
        self.assertEqual("++ foo\n", li.to_str(is_merge=False))

    def test_no_newline(self) -> None:
        li = Line((1, 1, 0, 0), "foo")
        self.assertEqual(
            "-- foo\n\\ no newline at end of file", li.to_str(is_merge=True)
        )
        self.assertEqual(
            "++ foo\n\\ no newline at end of file", li.to_str(is_merge=False)
        )


class HunkTest(TestCase):
    def test_context_hunk(self) -> None:
        li = Line((1, 1, 1, 1), "foo\n")
        h = Hunk(
            [li],
            (1, 1, 1, 1),
        )
        assert h.is_entirely_context
        assert h.lengths == (1, 1, 1, 1)
        assert h.to_str(is_merge=True) == ""

    def test_mixed_hunk(self) -> None:
        a = Line((0, 0, 0, 1), "foo\n")
        b = Line((1, 1, 1, 1), "bar\n")
        h = Hunk(
            [a, b],
            (1, 1, 1, 1),
        )
        assert not h.is_entirely_context
        assert h.lengths == (1, 1, 1, 2)
        self.assertEqual(
            """\
@@@@ -1,1 -1,1 -1,1 +1,2 @@@@
+++foo
   bar
""",
            h.to_str(is_merge=True),
        )

        self.assertEqual(
            """\
@@@@ -1,2 +1,1 +1,1 +1,1 @@@@
---foo
   bar
""",
            h.to_str(is_merge=False),
        )


class DivergentDiffTest(TestCase):
    def test_combined_diff_exceptions(self) -> None:
        with self.assertRaises(ValueError):
            combined_diff(["a", "b"], ["c", "d"])
        with self.assertRaises(ValueError):
            combined_diff([], [])

    def test_basic(self) -> None:
        result = combined_diff([""], ["a\n", "b\n"])
        self.assertEqual(
            """\
--- a/file
+++ b/file
+++ b/file
@@@ -1,0 +1,1 +1,1 @@@
+ a
 +b
""",
            result,
        )

    def test_basic_with_explicit_names(self) -> None:
        result = combined_diff(
            [""], ["a\n", "b\n"], from_filenames=["x"], to_filenames=["y", "z"]
        )
        self.assertEqual(
            """\
--- x
+++ y
+++ z
@@@ -1,0 +1,1 +1,1 @@@
+ a
 +b
""",
            result,
        )

    def test_basic_context(self) -> None:
        result = combined_diff(["a\nb\nc\nd\ne\nf\ng\n"], ["b\nc\nd\ne\nf\n"])
        self.assertEqual(
            """\
--- a/file
+++ b/file
@@ -1,7 +1,5 @@
-a
 b
 c
 d
 e
 f
-g
""",
            result,
        )

    def test_basic_context1(self) -> None:
        result = combined_diff(
            ["a\nb\nc\nd\ne\nf\ng\n"], ["b\nc\nd\ne\nf\n"], context=1
        )
        self.assertEqual(
            """\
--- a/file
+++ b/file
@@ -1,2 +1,1 @@
-a
 b
@@ -5,2 +4,1 @@
 f
-g
""",
            result,
        )

    def test_basic_context_final_output(self) -> None:
        result = combined_diff(["a\nb\nc\nd\n"], ["a\nb\nc\nd\n"], context=1)
        self.assertEqual(
            """\
--- a/file
+++ b/file
""",
            result,
        )

    def test_two(self) -> None:
        self.assertEqual(
            [
                Line((1, 1), "a\n"),
                Line((1, 0), "b\n"),
                Line((1, 0), "c\n"),
            ],
            _contributions(["a\n"], ["a\nb\nc\n"], False),
        )

    def test_contributions_exceptions(self) -> None:
        with self.assertRaises(ValueError):
            _contributions([], [], False)
        with self.assertRaises(ValueError):
            _contributions(["a", "b"], ["c", "d"], False)

    def test_three(self) -> None:
        self.assertEqual(
            [
                Line((1, 1, 1), "a\n"),
                Line((0, 1, 1), "b\n"),
                Line((0, 0, 1), "c\n"),
            ],
            _contributions(["a\nb\nc\n"], ["a\n", "a\nb\n"], False),
        )

    def test_three_insert_same(self) -> None:
        # this covers the path inside `helper` when we already have a matching line
        self.assertEqual(
            [
                Line((1, 1, 1), "a\n"),
                Line((1, 0, 1), "c\n"),
                Line((1, 1, 0), "b\n"),
                Line((0, 1, 0), "d\n"),
            ],
            _contributions(["a\nc\n"], ["a\nb\nc\n", "a\nb\nd\n"], False),
        )


class MergeDiffTest(TestCase):
    def test_not_actually_merge(self) -> None:
        result = combined_diff(["a\nc\n"], ["b\nc\n"])
        self.assertEqual(
            """\
--- a/file
+++ b/file
@@ -1,2 +1,2 @@
-a
+b
 c
""",
            result,
        )

    def test_basic(self) -> None:
        result = combined_diff(["a\n", "b\n"], [""])
        self.assertEqual(
            """\
--- a/file
--- a/file
+++ b/file
@@@ -1,1 -1,1 +1,0 @@@
- a
 -b
""",
            result,
        )

    def test_basic_with_explicit_names(self) -> None:
        result = combined_diff(
            ["a\n", "b\n"],
            [""],
            from_filenames=["x", "y"],
            to_filenames=["z"],
        )
        self.assertEqual(
            """\
--- x
--- y
+++ z
@@@ -1,1 -1,1 +1,0 @@@
- a
 -b
""",
            result,
        )

    def test_two(self) -> None:
        self.assertEqual(
            [
                Line((1, 1), "a\n"),
                Line((0, 1), "b\n"),
                Line((0, 1), "c\n"),
            ],
            _contributions(["a\n"], ["a\nb\nc\n"], True),
        )

    def test_two_removal(self) -> None:
        self.assertEqual(
            [
                Line((1, 1), "a\n"),
                Line((1, 0), "b\n"),
            ],
            _contributions(["a\nb\n"], ["a\n"], True),
        )

    def test_two_replace(self) -> None:
        self.assertEqual(
            [
                Line((1, 0), "a\n"),
                Line((0, 1), "x\n"),
                Line((1, 1), "b\n"),
            ],
            _contributions(["a\nb\n"], ["x\nb\n"], True),
        )

    def test_three(self) -> None:
        self.assertEqual(
            [
                Line((1, 1, 1), "a\n"),
                Line((0, 1, 1), "b\n"),
                Line((0, 0, 1), "c\n"),
            ],
            _contributions(["a\n", "a\nb\n"], ["a\nb\nc\n"], True),
        )

    def test_three_insert_same(self) -> None:
        # this covers the path inside `helper` when we already have a matching line
        self.assertEqual(
            [
                Line((1, 1, 1), "a\n"),
                Line((1, 1, 0), "b\n"),
                Line((0, 1, 0), "d\n"),
                Line((1, 0, 1), "c\n"),
            ],
            _contributions(["a\nb\nc\n", "a\nb\nd\n"], ["a\nc\n"], True),
        )



import unittest

from .click import ColorTest  # noqa: F401
from .combined import DivergentDiffTest, HunkTest, LineTest, MergeDiffTest  # noqa: F401
from .general import ParityTest  # noqa: F401
from .patch import PatchTest  # noqa: F401

unittest.main()



import subprocess
import tempfile
import unittest
from pathlib import Path

from parameterized import parameterized

from .. import unified_diff


class ParityTest(unittest.TestCase):
    @parameterized.expand(  # type: ignore
        [
            ("a", "a"),
            ("a", "b"),
            ("a\n", "b"),
            ("a", "b\n"),
            ("a\n", "b\n"),
        ]
    )
    def test_parity(self, a: str, b: str) -> None:
        with tempfile.TemporaryDirectory() as d:
            a_path = Path(d) / "a"
            a_path.mkdir()
            b_path = Path(d) / "b"
            b_path.mkdir()
            (a_path / "file").write_text(a)
            (b_path / "file").write_text(b)

            # Notably, diff exits 1 when the files are different :/
            # Force the labels because it would otherwise include timestamps.
            proc = subprocess.run(
                ["diff", "--label", "a/file", "--label", "b/file", "-u", "a", "b"],
                cwd=d,
                encoding="utf-8",
                stdout=subprocess.PIPE,
            )
            if "\n" in proc.stdout:
                expected = proc.stdout[proc.stdout.index("\n") + 1 :]
            else:
                expected = ""

            actual = unified_diff(a, b, "file")
            self.assertEqual(expected, actual)

    def test_absolute_paths(self) -> None:
        actual = unified_diff("a\n", "a\nb\n", "/file")
        self.assertEqual(
            """\
--- /file
+++ /file
@@ -1 +1,2 @@
 a
+b
""",
            actual,
        )



import random
import unittest
from typing import Any, List, Optional
from unittest.mock import patch

from parameterized import parameterized

from .. import unified_diff
from ..patch import (
    _context_match,
    _parse_position_line,
    _split_hunks,
    apply_single_file,
    PatchException,
)


class PatchTest(unittest.TestCase):
    @parameterized.expand(  # type: ignore
        [
            ("a", "b"),
            ("", "b"),
            ("a", ""),
            ("", "b\n"),
            ("a\n", ""),
            ("a\nb\n", "a\n"),
            ("a\nb\n", "b\n"),
            ("a\nb\n", "a\nb"),
            ("a\nb", "a\nb\n"),
        ]
    )
    def test_patch(self, a: str, b: str) -> None:
        diff = unified_diff(a, b, "foo")
        result = apply_single_file(a, diff)
        self.assertEqual(b, result)

        # Although we don't produce these, allow CRLF on the "No newline" line
        # to strip the full previous newline.
        if "No newline" in diff:
            dos_diff = diff.replace("\n\\ No newline", "\r\n\\ No newline")
            result = apply_single_file(a, dos_diff)
            self.assertEqual(b, result)

    @parameterized.expand(  # type: ignore
        [
            ("", "b\r\n"),
            ("a\r\n", ""),
            ("a\r\nb\r\n", "a\r\n"),
            ("a\r\nb\r\n", "b\r\n"),
            ("a\r\nb\r\n", "a\r\nb"),
            ("a\r\nb", "a\r\nb\r\n"),
        ]
    )
    def test_patch_crlf(self, a: str, b: str) -> None:
        diff = unified_diff(a, b, "foo")
        result = apply_single_file(a, diff)
        self.assertEqual(b, result)

    @parameterized.expand(  # type: ignore
        [
            (0,),
            (1,),
            (2,),
            (3,),
        ]
    )
    def test_exhaustive(self, context: int) -> None:
        for i in range(100):
            a = "".join(
                [random.choice(["a\n", "b\n", "c\n", "d\n"]) for x in range(10)]
            )
            b = "".join(
                [random.choice(["a\n", "b\n", "c\n", "d\n"]) for x in range(10)]
            )

            diff = unified_diff(a, b, "file", context)
            result = apply_single_file(a, diff)
            self.assertEqual(b, result)

    @parameterized.expand(  # type: ignore
        [
            ("@@ -5 +9 @@", [5, 1, 9, 1]),
            ("@@ -5,2 +9,3 @@", [5, 2, 9, 3]),
            ("@@ invalid @@", None),
        ]
    )
    def test_parse_position_line(
        self, line: str, expected: Optional[List[int]]
    ) -> None:
        if expected is None:
            with self.assertRaises(PatchException):
                _parse_position_line(line)
        else:
            self.assertEqual(expected, _parse_position_line(line))

    @parameterized.expand(  # type: ignore
        [
            ("---\n+++\n@@ -1 +1 @@\n-invalid\n", "Failed to apply with offset at 0"),
            ("---\n+++\n@@ -1 +1 @@\n invalid\n", "Failed to apply with offset at 0"),
            ("---\n+++\n@@ -1 +1 @@\nxinvalid\n", "Unknown line 'xinvalid\\\\n' at 0"),
        ]
    )
    def test_exceptions(self, diff: str, msg: str) -> None:
        with self.assertRaisesRegex(PatchException, msg):
            apply_single_file("foo\n", diff)

    @parameterized.expand(  # type: ignore
        [
            ("---\n+++\n@@ -1 +1 @@\n-invalid\n", "DELETE fail at 0"),
            ("---\n+++\n@@ -1 +1 @@\n invalid\n", "EQUAL fail at 0"),
            ("---\n+++\n@@ -1 +1 @@\nxinvalid\n", "Unknown line 'xinvalid\\\\n' at 0"),
        ]
    )
    def test_exceptions_no_offset(self, diff: str, msg: str) -> None:
        with self.assertRaisesRegex(PatchException, msg):
            apply_single_file("foo\n", diff, allow_offsets=False)

    def test_split_hunks_edge_cases(self) -> None:
        with self.assertRaisesRegex(PatchException, "Lines without hunk header.*"):
            _split_hunks(["foo\n"])
        self.assertEqual([], _split_hunks([]))

    @patch("moreorless.patch.LOG.info")
    def test_patch_small_offset(self, log_info: Any) -> None:
        a = "a\nb\nc\n"
        b = "a\nB\nc\n"
        modified = "x\n" + a
        expected = "x\n" + b

        diff = unified_diff(a, b, "foo")
        result = apply_single_file(modified, diff)
        self.assertEqual(expected, result)
        log_info.assert_called_with("Offset 1")

    @parameterized.expand(  # type: ignore
        [
            ((["0", "1", "2", "3"], 0, 5, 0), 0),  # can match at start
            ((["0", "1", "2", "3"], 0, 5, 1), 0),  # can match earlier
            ((["1", "2", "3", "4"], 0, 5, 0), 1),  # can match later
            ((["4"], 0, 5, 0), 4),  # can match later
            ((["5"], 0, 4, 3), None),  # no possible match, starts past mid
        ]
    )
    def test_context_match(self, args: Any, expected: Optional[int]) -> None:
        self.assertEqual(expected, _context_match(["0", "1", "2", "3", "4"], *args))

    def test_context_match_tie(self) -> None:
        # ties resolve earlier
        self.assertEqual(0, _context_match(["0", "1", "0"], ["0"], 0, 3, 1))

    def test_edge_cases(self) -> None:
        with self.assertRaisesRegex(PatchException, "negative range_start"):
            _context_match(["0", "1", "2"], ["0"], -1, 3, 0)
        with self.assertRaisesRegex(PatchException, "flipped range"):
            _context_match(["0", "1", "2"], ["0"], 3, 0, 0)
        with self.assertRaisesRegex(PatchException, "past end"):
            _context_match(["0", "1", "2"], ["0"], 0, 4, 0)
        with self.assertRaisesRegex(PatchException, "start before range_start"):
            _context_match(["0", "1", "2"], ["0"], 1, 3, 0)
        with self.assertRaisesRegex(PatchException, "start past range_end"):
            _context_match(["0", "1", "2"], ["0"], 0, 3, 3)



import unittest
from typing import Any
from unittest.mock import call, patch

from .. import unified_diff
from ..click import echo_color_precomputed_diff, echo_color_unified_diff


class ColorTest(unittest.TestCase):
    @patch("click.secho")
    def test_echo_color_unified_diff(self, secho: Any) -> None:
        echo_color_unified_diff("a\nb\n", "a\nc\n", "x")
        secho.assert_has_calls(
            [
                call("--- a/x\n", bold=True, nl=False),
                call("+++ b/x\n", bold=True, nl=False),
                call("@@ -1,2 +1,2 @@\n", fg="cyan", nl=False),
                call(" a\n", nl=False),
                call("-b\n", fg="red", nl=False),
                call("+c\n", fg="green", nl=False),
            ]
        )

    @patch("click.secho")
    def test_echo_color_precomputed_diff(self, secho: Any) -> None:
        diff = unified_diff("a\nb\n", "a\nc\n", "x")
        echo_color_precomputed_diff(diff)
        secho.assert_has_calls(
            [
                call("--- a/x\n", bold=True, nl=False),
                call("+++ b/x\n", bold=True, nl=False),
                call("@@ -1,2 +1,2 @@\n", fg="cyan", nl=False),
                call(" a\n", nl=False),
                call("-b\n", fg="red", nl=False),
                call("+c\n", fg="green", nl=False),
            ]
        )
